/* DIFFH - A diff frontend for output in HTML format.
 * Module:  OUTPUT
 * Author:  Matt Giuca
 * Created: 6/2/06
 * Updated: 15/4/06
 *
 * Copyright (C) 2006 Matt Giuca <mattgiuca@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either 
 * version 2 of the License, or (at your option) any later 
 * version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* Module: OUTPUT
 * This module takes UDiff structures containing complete diff output, parses
 * them, and rewrites them in HTML format.
 */

#include <ctype.h>
#include "global.h"
#include "output.h"
#include "chunk.h"

/* PRIVATE FUNCTION DECLARATIONS */
static void print_html_prologue(char* from_filename, char* to_filename);
static void print_html_epilogue(void);
static void print_headers(UDiff* diff);
static void hunk_process(Hunk* hunk);
static void print_hunk_header(int from_line_start, int to_line_start);
static void print_unchanged(char* line);
static char* convert_to_valid_html(char* str);

/* MODULE FUNCTION CODE */

/* Function: diff_process
 * Takes a unified diff, fully processes it, outputs to HTML.
 * This function outputs the entire diff <table> element to HTML.
 * Precond: Global outstream must be valid.
 */
void diff_process(UDiff* diff)
{
    int i;

    /* Open HTML, table */
    print_html_prologue(diff->from_filename, diff->to_filename);

    /* Print file name/stamp headers */
    print_headers(diff);

    /* Print hunks */
    for (i=0; i<diff->numhunks; i++)
        hunk_process(diff->hunk[i]);

    /* Close table, HTML */
    print_html_epilogue();
}


/* Function: print_html_prologue
 * Prints the HTML source from beginning up to and including the opening
 * <table> tag.
 */
void print_html_prologue(char* from_filename, char* to_filename)
{
    fputs("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" "
        "\"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">" NL, outstream);
    fputs("<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\"> "
        NL, outstream);
    fputs(HINDH1 "<head>" NL, outstream);
    fputs(HINDH2 "<title>", outstream);
    if (strcmp(from_filename, to_filename))
        fprintf(outstream, PAGETITLE_FORMAT_TWO, from_filename, to_filename);
    else
        fprintf(outstream, PAGETITLE_FORMAT_ONE, from_filename);
    fputs("</title>" NL, outstream);
    fputs(HINDH2 "<meta http-equiv=\"Content-Type\" content=\"text/html; "
        "charset=UTF-8\" />" NL HINDH2
        "<meta http-equiv=\"Content-Style-Type\" content=\"text/css\" />"
        NL, outstream);
    if (custom_css_filename)
        fprintf(outstream, HINDH2 "<link rel=\"stylesheet\" type=\"text/css\" "
            "href=\"%s\" />" NL, custom_css_filename);
    else
    {
        fputs(HINDH2 "<style type=\"text/css\">" NL HINDH2 "<!--" NL,
            outstream);

        /* BEGIN CSS SECTION */
        fprintf(outstream, HINDS1 "body {" NL
                HINDS2 "font-family: %s;" NL
                HINDS2 "font-size: %s;" NL
                HINDS2 "background: %s;" NL
                HINDS1 "}" NL, css_bodyfont, css_bodyfontsize,
            css_bodybackcolor);

        fputs(HINDS1 "div#content {" NL
                HINDS2 "font-size: 127%;" NL
                HINDS1 "}" NL, outstream);

        fputs(HINDS1 "table.diff-table, td.diff-heading {" NL
                HINDS2 "border: 0;" NL
                HINDS1 "}" NL, outstream);

        fputs(HINDS1 "td {" NL
                HINDS2 "text-align: center;" NL
                HINDS1 "}" NL, outstream);

        fputs(HINDS1 "td.diff-heading {" NL
                HINDS2 "width: 50%;" NL
                HINDS1 "}" NL, outstream);

        fputs(HINDS1 "td.diff-linenum {" NL
                HINDS2 "text-align: left;" NL
                HINDS2 "font-weight: bold;" NL
                HINDS1 "}" NL, outstream);

        fputs(HINDS1 "td.diff-added, td.diff-deleted, td.diff-unchanged {" NL
                HINDS2 "text-align: left;" NL
                HINDS2 "font-size: 85%;" NL
                HINDS1 "}" NL, outstream);

        fprintf(outstream, HINDS1 "td.diff-added {" NL
                HINDS2 "background-color: %s;" NL
                HINDS1 "}" NL, css_color_added);

        fprintf(outstream, HINDS1 "td.diff-deleted {" NL
                HINDS2 "background-color: %s;" NL
                HINDS1 "}" NL, css_color_deleted);

        fprintf(outstream, HINDS1 "td.diff-unchanged {" NL
                HINDS2 "background-color: %s;" NL
                HINDS1 "}" NL, css_color_unchanged);

        fprintf(outstream, HINDS1 "span.diff-d, span.diff-a {" NL
                HINDS2 "color: %s;" NL
                HINDS2 "font-weight: bold;" NL
                HINDS1 "}" NL, css_color_change);

        /* END CSS SECTION */
        fputs(HINDH2 "-->" NL HINDH2 "</style>" NL, outstream);
    }

    fputs(HINDH1 "</head>" NL NL, outstream);
    fputs(HINDB1 "<body><div id=\"content\">" NL, outstream);
    fputs(HINDB2 TABLE_START NL, outstream);
}

/* Function: print_html_epilogue
 * Prints the HTML source from and including the closing </table> tag, up to
 * the end of the HTML document.
 */
void print_html_epilogue(void)
{
    fputs(HINDB2 TABLE_END NL, outstream);
    fputs(HINDB2 "<p><small>" FOOTER_TEXT "</small></p>" NL, outstream);
    fputs(HINDB1 "</div></body>" NL, outstream);
    fputs("</html>" NL, outstream);
}

/* Function: print_headers
 * Takes a unified diff and prints out its header information.
 */
void print_headers(UDiff* diff)
{
    fputs(HINDT1 TABLE_ROWSTART NL HINDT2, outstream);
    fputs(TABLE_CELLSTART_DIFFHEAD, outstream);

    if (diff->from_version && *diff->from_version)
        fprintf(outstream, DIFFHEAD_FORMAT_NAME_VER_TIME,
            diff->from_filename, diff->from_version, diff->from_time);
    else
        fprintf(outstream, DIFFHEAD_FORMAT_NAME_TIME,
            diff->from_filename, diff->from_time);

    fputs(TABLE_CELLEND, outstream);
    fputs(NL HINDT2, outstream);
    fputs(TABLE_CELLSTART_DIFFHEAD, outstream);

    if (diff->to_version && *diff->to_version)
        fprintf(outstream, DIFFHEAD_FORMAT_NAME_VER_TIME,
            diff->to_filename, diff->to_version, diff->to_time);
    else
        fprintf(outstream, DIFFHEAD_FORMAT_NAME_TIME,
            diff->to_filename, diff->to_time);

    fputs(TABLE_CELLEND, outstream);
    fputs(NL HINDT1 TABLE_ROWEND NL, outstream);
}

/* Function: hunk_process
 * Takes a hunk, fully processes it, outputs to HTML.
 * Precond: Global outstream must be valid.
 */
void hunk_process(Hunk* hunk)
{
    int i, j;
    int numlines = hunk->numlines;
    int numdeleted, numadded;
    char** deleted, ** added;

    /* Print the hunk header row */
    print_hunk_header(hunk->from_line_start, hunk->to_line_start);

    /* Process all lines of the hunk */
    for (i=0; i<numlines; i++)
    {
        if (hunk->line_type[i] == LINE_UNCHANGED)
            /* Output this as an unchanged line */
            print_unchanged(hunk->line[i]);
        else
        {
            /* Read all additions/deletions before the next unchanged, and
                parse them together as a chunk */
            /* First pass - count numbers */
            numdeleted = numadded = 0;
            for (j=i; j<numlines && hunk->line_type[j] != LINE_UNCHANGED;
                j++)
            {
                if (hunk->line_type[j] == LINE_ADDED)
                    numadded++;
                else
                    numdeleted++;
            }
            deleted = malloc(numdeleted * sizeof(char*));
            added = malloc(numadded * sizeof(char*));
            numdeleted = numadded = 0;
            /* Second pass - collect lines into arrays */
            for (; i<numlines && hunk->line_type[i] != LINE_UNCHANGED; i++)
            {
                if (hunk->line_type[i] == LINE_ADDED)
                    added[numadded++] = hunk->line[i];
                else
                    deleted[numdeleted++] = hunk->line[i];
            }
            i--;
            /* Pass to chunk module to print */
            chunk_process(numdeleted, deleted, numadded, added);

            free(deleted);
            free(added);
        }
    }
}

/* Function: print_hunk_header
 * Takes the start lines of each side of the hunk, and prints out the
 * hunk's header row in HTML format.
 */
void print_hunk_header(int from_line_start, int to_line_start)
{
    fputs(HINDT1 TABLE_ROWSTART NL HINDT2, outstream);
    fputs(TABLE_CELLSTART_HUNKHEAD, outstream);
    fprintf(outstream, LINENUM_FORMAT, from_line_start);
    fputs(TABLE_CELLEND, outstream);
    fputs(NL HINDT2, outstream);
    fputs(TABLE_CELLSTART_HUNKHEAD, outstream);
    fprintf(outstream, LINENUM_FORMAT, to_line_start);
    fputs(TABLE_CELLEND, outstream);
    fputs(NL HINDT1 TABLE_ROWEND NL, outstream);
}

/* Function: print_unchanged
 * Takes a single line and prints it to HTML as an unchanged line.
 * This prints table cells, with two duplicate copies of the line.
 */
void print_unchanged(char* line)
{
    line = convert_to_valid_html(line);
    fputs(HINDT1 TABLE_ROWSTART NL HINDT2, outstream);
    if (*line == '\0')
        /* Need &nbsp; to make agent display the row */
        fputs(TABLE_CELL_UNCHANGED_EMPTY, outstream);
    else
        fputs(TABLE_CELL_UNCHANGED, outstream);
    fputs(TABLE_CELLSTART_UNCHANGED, outstream);
    fputs(line, outstream);
    fputs(TABLE_CELLEND, outstream);
    fputs(NL HINDT2, outstream);
    fputs(TABLE_CELL_UNCHANGED TABLE_CELLSTART_UNCHANGED, outstream);
    fputs(line, outstream);
    fputs(TABLE_CELLEND, outstream);
    fputs(NL HINDT1 TABLE_ROWEND NL, outstream);
    free(line);
}

/* Function: convert_to_valid_html
 * Takes a string and changes any necessary characters into valid HTML
 * markup. Returns a newly-allocated string which must be freed.
 */
char* convert_to_valid_html(char* str)
{
    char* i;
    int t;
    char* ret = malloc(DEFAULT_MALLOC_BIG * sizeof(char));
    size_t retsize = 0, retsize_new = 0;
    unsigned int malloc_size = DEFAULT_MALLOC_BIG;
    char* thisword;
    char thisword_mallocd;  /* boolean */
    int numspaces, numspaces_d2;

    for (i=str; *i; i++)
    {
        thisword = NULL;
        thisword_mallocd = FALSE;
        switch (*i)
        {
            case ' ':   /* Spaces - if multiple, replace with &nbsp; */
            case '\t':
            if (*i == ' ')
            {   /* Read in all spaces to numspaces, set i at last space */
                numspaces = 0;
                for (; *i == ' '; i++)
                    numspaces++;
                i--;
                if (numspaces == 1) break;
            }
            else
                numspaces = tabwidth;
            numspaces_d2 = numspaces / 2;
            thisword = malloc((7*numspaces_d2 + 6*(numspaces%2) + 1)
                * sizeof(char));
            for (t=0; t<numspaces_d2; t++)
                strncpy(thisword+(t*7), HT_NBSP " ", 7);
            if (numspaces%2) strncpy(thisword+(t*7), HT_NBSP, 6);
            thisword[7*(numspaces/2) + 6*(numspaces%2)] = '\0';
            thisword_mallocd = TRUE;
            break;
            case '&':
            thisword = HT_AMP;
            break;
            case '\"':
            thisword = HT_QUOT;
            break;
            case '<':
            thisword = HT_LT;
            break;
            case '>':
            thisword = HT_GT;
            break;
        }
        if (thisword)
        {
            retsize_new += strlen(thisword);
            if (retsize_new+1 >= malloc_size)
                ret = realloc(ret, (malloc_size*=2) * sizeof(char));
            strcpy(ret+retsize, thisword);
            if (thisword_mallocd) free(thisword);
        }
        else
        {
            retsize_new++;
            if (retsize_new+1 >= malloc_size)
                ret = realloc(ret, (malloc_size*=2) * sizeof(char));
            ret[retsize] = *i;
        }
        retsize = retsize_new;
    }
    ret[retsize] = '\0';
    return ret;
}
